%% DIExample2
% Generate Random DT Examples

%load Ex1;

%% Nominal System
Ag = -0.5;
Bg = [0.5 0.4];
Cg = [2.5; 2];
Dg = [0 0.6; 0 0.9];
Ts = 1;

G = ss(Ag,Bg,Cg,Dg,Ts);

Ts = G.Ts;
G.InputName = {'w'; 'd'};
G.OutputName = {'v'; 'e'};

%% Multipliers 
% (Psi1Hat,J1Hat) is a J-spectral factorization for Psi1' M1 Psi1.
M1 = [1 0; 0 -1];

Apsi1 = -0.3;
Bpsi1 = [1.3 0];
Cpsi1 = [0; -0.1];
Dpsi1 = [0.2 0; 0 -0.1];
Psi1 = ss(Apsi1,Bpsi1,Cpsi1,Dpsi1,Ts);
Pi1 = Psi1'*M1*Psi1;


M2 = [1 0; 0 -1];
Psi2 = [-0.5 0.3; 0 1.7];
Psi2 = ss([],[],[],Psi2,Ts);
Pi2 = Psi2'*M2*Psi2;

% "Hard" IQCs
% [Psi1Hat,J1Hat]=jspec(Pi1);
Psi1Hat=Psi1;
J1Hat=M1;

Psi2Hat = Psi2;
J2Hat = M2;

%% Standard Dissipation Inequality Analysis

% Form extended system using "hard" IQCs 
% Note: The individual IQCs Psi1Hat and Psi2Hat are hard but the
% vertical concantenation Psi is not square and not hard.
Psi = [Psi1Hat; Psi2Hat];
Psi.InputName = {'v'; 'w'};
Psi.OutputName = {'r'};
ExtSys = connect(G,Psi,{'w';'d'},{'r';'e'});

A = ExtSys.A;
B1 = ExtSys.B(:,1);
B2 = ExtSys.B(:,2);
C1 = ExtSys.C(1:end-1,:);
C2 = ExtSys.C(end,:);
D11 = ExtSys.D(1:end-1,1:end-1);
D12 = ExtSys.D(1:end-1,end);
D21 = ExtSys.D(end,1:end-1);
D22 = ExtSys.D(end,end);

% Dimensions
nx = size(A,1);
nd = 1;
ne = 1;
nw = 1;
nv = 1;

% Solve Standard DI LMI 
cvx_begin sdp quiet
    variables lam1 lam2 gamsq
    variable P(nx,nx) semidefinite;    
    M = blkdiag(lam1*J1Hat,lam2*J2Hat);
    
    minimize(gamsq)
    subject to
    [A B1 B2]'*P*[A B1 B2] - blkdiag(P,zeros(nw),gamsq*eye(nd)) ...
        + [C2 D21 D22]'*[C2 D21 D22] ...
        + [C1 D11 D12]'*M*[C1 D11 D12] <= 0;
        lam1>=0;
        lam2>=0;
        gamsq>=0;
cvx_end

% Verify Solution
LMI = [A B1 B2]'*P*[A B1 B2] - blkdiag(P,zeros(nw),gamsq*eye(nd)) ...
        + [C2 D21 D22]'*[C2 D21 D22] ...
        + [C1 D11 D12]'*M*[C1 D11 D12];

fprintf('\n -----Standard Dissipation Inequality Results');
fprintf('\n Gain = %4.3f',sqrt(gamsq));
fprintf('\n min(eig(P)) = %4.3f',min(real(eig(P))));
fprintf('\t max(eig(LMI)) = %4.3f\n',max(real(eig(LMI))));

% Store optimal variables for standard DI
gamsqS = gamsq;
PS = P;
lam1S = lam1;
lam2S = lam2;


%% Modified Dissipation Inequality Analysis

% Form extended system 
Psi = [Psi1; Psi2];
Psi.InputName = {'v'; 'w'};
Psi.OutputName = {'r'};
ExtSys = connect(G,Psi,{'w';'d'},{'r';'e'});

A = ExtSys.A;
B1 = ExtSys.B(:,1);
B2 = ExtSys.B(:,2);
C1 = ExtSys.C(1:end-1,:);
C2 = ExtSys.C(end,:);
D11 = ExtSys.D(1:end-1,1:end-1);
D12 = ExtSys.D(1:end-1,end);
D21 = ExtSys.D(end,1:end-1);
D22 = ExtSys.D(end,end);

% Solve Standard DI LMI 
cvx_begin sdp quiet
    variables lam1 lam2 gamsq 
    variable P(nx,nx) symmetric;    
    M = blkdiag(lam1*M1,lam2*M2);
    
    minimize(gamsq)
    subject to
    [A B1 B2]'*P*[A B1 B2] - blkdiag(P,zeros(nw),gamsq*eye(nd)) ...
        + [C2 D21 D22]'*[C2 D21 D22] ...
        + [C1 D11 D12]'*M*[C1 D11 D12] <= 0
    lam1>=0;
    lam2>=0;
    gamsq>=0;
cvx_end

% Verify Solution
LMI = [A B1 B2]'*P*[A B1 B2] - blkdiag(P,zeros(nw),gamsq*eye(nd)) ...
        + [C2 D21 D22]'*[C2 D21 D22] ...
        + [C1 D11 D12]'*M*[C1 D11 D12];

fprintf('\n -----Modified Dissipation Inequality Results');
fprintf('\n Gain = %4.3f',sqrt(gamsq));
fprintf('\n min(eig(P)) = %4.3f',min(real(eig(P))));
fprintf('\t max(eig(LMI)) = %4.3f\n',max(real(eig(LMI))));

% Store optimal variables for modified DI
gamsqM = gamsq;
PM = P;
lam1M = lam1;
lam2M = lam2;

%% Verify J-spectral Factorization Solution

% Solve DARE 
Apsi = Psi.A;
Bpsi = Psi.B;
Cpsi = Psi.C;
Dpsi = Psi.D;
Q = Cpsi'*M*Cpsi;
S = Cpsi'*M*Dpsi;
R = Dpsi'*M*Dpsi;
XHat = dare(Apsi,Bpsi,Q,R,S);

% Construct J-spectral Factorization
[U,T] = schur(R+Bpsi'*XHat*Bpsi);
ev = real( diag(T) );
pidx = find( ev>0 );
qidx = find( ev<0 );
Dpsi = diag( sqrt([ev(pidx); -ev(qidx)]) );
Dpsi = Dpsi*U(:,[pidx;qidx])';

p = length(pidx);
q = length(qidx);
JHat = blkdiag( eye(p), -eye(q) );

Cpsi = JHat*inv(Dpsi')*(Bpsi'*XHat*Apsi+S');
PsiHat = ss(Apsi,Bpsi,Cpsi,Dpsi,Ts);

% Verify J-spectral Factorization
PiLam = lam1*Pi1+lam2*Pi2;
E=norm( PiLam - PsiHat'*JHat*PsiHat,inf);

fprintf('\n ----- J-Spectral Factorization');
fprintf(' \n norm( Pi(lambda) - PsiHat~ JHat PsiHat ) = %4.3f', E);

fprintf('\n abs(pole(PsiHat)) = %4.3f \t isstable = %d',...
    max(abs(pole(PsiHat))), isstable(PsiHat) );

fprintf('\n abs(pole(PsiHat^(-1))) = %4.3f \t isstable = %d\n',...
    max(abs(pole(inv(PsiHat)))), isstable(inv(PsiHat)) );

% Transformed LMI
PsiHat.InputName = {'v'; 'w'};
PsiHat.OutputName = {'r'};
ExtSys = connect(G,PsiHat,{'w';'d'},{'r';'e'});

A = ExtSys.A;
B1 = ExtSys.B(:,1);
B2 = ExtSys.B(:,2);
C1 = ExtSys.C(1:end-1,:);
C2 = ExtSys.C(end,:);
D11 = ExtSys.D(1:end-1,1:end-1);
D12 = ExtSys.D(1:end-1,end);
D21 = ExtSys.D(end,1:end-1);
D22 = ExtSys.D(end,end);

PHat = P - [0 0;0 XHat];
gamsqHat = gamsqM;

LMIHat = [A B1 B2]'*PHat*[A B1 B2] - blkdiag(PHat,zeros(nw),gamsqHat*eye(nd)) ...
        + [C2 D21 D22]'*[C2 D21 D22] ...
        + [C1 D11 D12]'*JHat*[C1 D11 D12];

% Verify J-spectral Factorization LMI
fprintf('\n ----- LMI with J-Spectral Factorization');
fprintf('\n Gain = %4.3f',sqrt(gamsqHat));
fprintf('\n min(eig(P)) = %4.3f',min(real(eig(PHat))));
fprintf('\t max(eig(LMI)) = %4.3f\n',max(real(eig(LMIHat))));
